package com.wechi.adweb.bridge.google.ads.controller;

import com.wechi.adweb.bridge.common.APIStatus;
import com.wechi.adweb.bridge.common.BaseController;
import com.wechi.adweb.bridge.common.OpenAPIRequest;
import com.wechi.adweb.bridge.common.OpenAPIResponse;
import com.wechi.adweb.bridge.exception.BadRequestException;
import com.wechi.adweb.bridge.exception.DataException;
import com.wechi.adweb.bridge.google.ads.dto.CampaignStatsDTO;
import com.wechi.adweb.bridge.google.ads.dto.CustomerStatsDTO;
import com.wechi.adweb.bridge.google.ads.dto.MetricsDTO;
import com.wechi.adweb.bridge.google.ads.dto.ReportRequestDTO;
import com.wechi.adweb.bridge.google.ads.service.GoogleAdsService;
import com.wechi.adweb.bridge.util.JsonUtils;

import io.swagger.v3.oas.annotations.Operation;

import lombok.extern.slf4j.Slf4j;

import org.apache.commons.lang3.StringUtils;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.web.bind.annotation.*;

import java.util.List;
import java.util.Map;

/**
 * @author wfansh
 */
@Slf4j
@RestController
@RequestMapping("/api/google/ads")
public class GoogleAdsController extends BaseController {

    @Autowired private GoogleAdsService googleAdsService;

    @Operation(summary = "Get customer stats by customer ID.")
    @PostMapping("/customerStats")
    @ResponseBody
    public OpenAPIResponse<CustomerStatsDTO> getCustomerStats(
            @RequestBody OpenAPIRequest<ReportRequestDTO> apiRequest)
            throws BadRequestException, DataException {
        long start = System.currentTimeMillis();
        log.info("****** getCustomerStats() ****** apiRequest = {}", JsonUtils.toJson(apiRequest));
        ReportRequestDTO reportRequest = apiRequest.getData();

        // 1. Validates the request parameters.
        if (StringUtils.isEmpty(reportRequest.getCustomerId())) {
            throw new BadRequestException(apiRequest);
        }

        // 2. Executes the API request.
        CustomerStatsDTO customerStats = googleAdsService.getCustomerStats(reportRequest);
        log.info(
                "****** getCustomerStats() ****** duration = {} seconds", getElapsedSeconds(start));
        return OpenAPIResponse.<CustomerStatsDTO>builder()
                .status(APIStatus.SUCCESS)
                .data(customerStats)
                .build();
    }

    @PostMapping("/dailyStats")
    @ResponseBody
    public OpenAPIResponse<Map<String, MetricsDTO>> getDailyStats(
            @RequestBody OpenAPIRequest<ReportRequestDTO> apiRequest)
            throws BadRequestException, DataException {
        long start = System.currentTimeMillis();
        log.info("****** getDailyStats() ****** apiRequest = {}", JsonUtils.toJson(apiRequest));
        ReportRequestDTO reportRequest = apiRequest.getData();

        // 1. Validates the request parameters.
        if (!validateReportRequest(reportRequest)) {
            throw new BadRequestException(apiRequest);
        }

        // 2. Executes the API request.
        Map<String, MetricsDTO> dailyStats = googleAdsService.getDailyStats(reportRequest);
        log.info("****** getDailyStats() ****** duration = {} seconds", getElapsedSeconds(start));
        return OpenAPIResponse.<Map<String, MetricsDTO>>builder()
                .status(APIStatus.SUCCESS)
                .data(dailyStats)
                .build();
    }

    @PostMapping("/campaignStats")
    @ResponseBody
    public OpenAPIResponse<List<CampaignStatsDTO>> getCampaignStats(
            @RequestBody OpenAPIRequest<ReportRequestDTO> apiRequest)
            throws BadRequestException, DataException {
        long start = System.currentTimeMillis();
        log.info("****** getCampaignStats() ****** apiRequest = {}", JsonUtils.toJson(apiRequest));
        ReportRequestDTO reportRequest = apiRequest.getData();

        // 1. Validates the request parameters.
        if (!validateReportRequest(reportRequest)) {
            throw new BadRequestException(apiRequest);
        }

        // 2. Executes the API request.
        List<CampaignStatsDTO> campaignStats = googleAdsService.getCampaignStats(reportRequest);
        log.info(
                "****** getCampaignStats() ****** duration = {} seconds", getElapsedSeconds(start));
        return OpenAPIResponse.<List<CampaignStatsDTO>>builder()
                .status(APIStatus.SUCCESS)
                .data(campaignStats)
                .build();
    }

    @PostMapping("/keywordStats")
    @ResponseBody
    public OpenAPIResponse<Map<String, MetricsDTO>> getKeywordStats(
            @RequestBody OpenAPIRequest<ReportRequestDTO> apiRequest)
            throws BadRequestException, DataException {
        long start = System.currentTimeMillis();
        log.info("****** getKeywordStats() ****** apiRequest = {}", JsonUtils.toJson(apiRequest));
        ReportRequestDTO reportRequest = apiRequest.getData();

        // 1. Validates the request parameters.
        if (!validateReportRequest(reportRequest)) {
            throw new BadRequestException(apiRequest);
        }

        // 2. Executes the API request.
        Map<String, MetricsDTO> keywordStats = googleAdsService.getKeywordStats(reportRequest);
        log.info("****** getKeywordStats() ****** duration = {} seconds", getElapsedSeconds(start));
        return OpenAPIResponse.<Map<String, MetricsDTO>>builder()
                .status(APIStatus.SUCCESS)
                .data(keywordStats)
                .build();
    }

    @PostMapping("/placementStats")
    @ResponseBody
    public OpenAPIResponse<Map<String, MetricsDTO>> getPlacementStats(
            @RequestBody OpenAPIRequest<ReportRequestDTO> apiRequest)
            throws BadRequestException, DataException {
        long start = System.currentTimeMillis();
        log.info("****** getPlacementStats() ****** apiRequest = {}", JsonUtils.toJson(apiRequest));
        ReportRequestDTO reportRequest = apiRequest.getData();

        // 1. Validates the request parameters.
        if (!validateReportRequest(reportRequest)) {
            throw new BadRequestException(apiRequest);
        }

        // 2. Executes the API request.
        Map<String, MetricsDTO> placementStats = googleAdsService.getPlacementStats(reportRequest);
        log.info(
                "****** getPlacementStats() ****** duration = {} seconds",
                getElapsedSeconds(start));
        return OpenAPIResponse.<Map<String, MetricsDTO>>builder()
                .status(APIStatus.SUCCESS)
                .data(placementStats)
                .build();
    }

    @PostMapping("/countryStats")
    @ResponseBody
    public OpenAPIResponse<Map<String, MetricsDTO>> getCountryStats(
            @RequestBody OpenAPIRequest<ReportRequestDTO> apiRequest)
            throws BadRequestException, DataException {
        long start = System.currentTimeMillis();
        log.info("****** getCountryStats() ****** apiRequest = {}", JsonUtils.toJson(apiRequest));
        ReportRequestDTO reportRequest = apiRequest.getData();

        // 1. Validates the request parameters.
        if (!validateReportRequest(reportRequest)) {
            throw new BadRequestException(apiRequest);
        }

        // 2. Executes the API request.
        Map<String, MetricsDTO> countryStats = googleAdsService.getCountryStats(reportRequest);
        log.info("****** getCountryStats() ****** duration = {} seconds", getElapsedSeconds(start));
        return OpenAPIResponse.<Map<String, MetricsDTO>>builder()
                .status(APIStatus.SUCCESS)
                .data(countryStats)
                .build();
    }

    private boolean validateReportRequest(ReportRequestDTO reportRequest) {
        return StringUtils.isNotEmpty(reportRequest.getCustomerId())
                && StringUtils.isNotEmpty(reportRequest.getStartDate())
                && StringUtils.isNotEmpty(reportRequest.getEndDate());
    }
}
