package com.wechi.adweb.bridge.google.gtm.service;

import static com.wechi.adweb.bridge.google.common.ResourceDTO.parseId;

import com.google.api.client.googleapis.javanet.GoogleNetHttpTransport;
import com.google.api.client.json.gson.GsonFactory;
import com.google.api.client.util.Lists;
import com.google.api.services.tagmanager.TagManager;
import com.google.api.services.tagmanager.TagManagerScopes;
import com.google.api.services.tagmanager.model.*;
import com.google.auth.http.HttpCredentialsAdapter;
import com.google.auth.oauth2.GoogleCredentials;
import com.wechi.adweb.bridge.exception.DataException;
import com.wechi.adweb.bridge.google.gtm.dto.*;
import com.wechi.adweb.bridge.matomo.service.MatomoTagService;
import com.wechi.adweb.bridge.util.JsonUtils;

import lombok.extern.slf4j.Slf4j;

import org.apache.commons.lang3.BooleanUtils;
import org.apache.commons.lang3.StringUtils;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.scheduling.annotation.Async;
import org.springframework.stereotype.Service;

import java.io.IOException;
import java.security.GeneralSecurityException;
import java.time.LocalDateTime;
import java.time.format.DateTimeFormatter;
import java.util.Collections;
import java.util.List;
import java.util.Optional;

import javax.annotation.PostConstruct;

/**
 * @author wfansh
 */
@Slf4j
@Service
public class GTMService {

    @Value("${google.service.account.key}")
    private String serviceAccountKey;

    @Value("${google.gtm.container.admins}")
    private List<String> containerAdmins;

    @Autowired private SnippetService snippetService;

    @Autowired private MatomoTagService matomoTagService;

    public TagManager tagManager;

    @PostConstruct
    private void init() throws GeneralSecurityException, IOException {
        // Service account authorization;
        GoogleCredentials credentials =
                GoogleCredentials.fromStream(
                                this.getClass()
                                        .getClassLoader()
                                        .getResourceAsStream(serviceAccountKey))
                        .createScoped(TagManagerScopes.all());

        this.tagManager =
                new TagManager.Builder(
                                GoogleNetHttpTransport.newTrustedTransport(),
                                GsonFactory.getDefaultInstance(),
                                new HttpCredentialsAdapter(credentials))
                        .build();
    }

    public List<GTMAccountDTO> listAccounts() throws DataException {
        try {
            List<GTMAccountDTO> gtmAccounts = Lists.newArrayList();
            for (Account account :
                    Optional.ofNullable(tagManager.accounts().list().execute().getAccount())
                            .orElse(List.of())) {
                GTMAccountDTO gtmAccount =
                        GTMAccountDTO.builder()
                                .id(account.getAccountId())
                                .resourceName(account.getPath())
                                .displayName(account.getName())
                                .supportUserPermissions(
                                        account.getFeatures().getSupportUserPermissions())
                                .supportMultipleContainers(
                                        account.getFeatures().getSupportMultipleContainers())
                                .build();

                // Adds into the list;
                gtmAccounts.add(gtmAccount);
            }

            log.info("listAccounts : {}", JsonUtils.toJson(gtmAccounts));
            return gtmAccounts;
        } catch (IOException e) {
            log.error(e.getMessage());
            throw new DataException(e);
        }
    }

    public List<GTMContainerDTO> listContainers(String accountResourceName) throws DataException {
        try {
            List<GTMContainerDTO> gtmContainers = Lists.newArrayList();
            for (Container container :
                    Optional.ofNullable(
                                    tagManager
                                            .accounts()
                                            .containers()
                                            .list(accountResourceName)
                                            .execute()
                                            .getContainer())
                            .orElse(List.of())) {
                GTMContainerDTO gtmContainer =
                        GTMContainerDTO.builder()
                                .id(container.getContainerId())
                                .resourceName(container.getPath())
                                .displayName(container.getName())
                                .accountId(container.getAccountId())
                                .domainNames(container.getDomainName())
                                .publicId(container.getPublicId())
                                .tagIds(container.getTagIds())
                                .notes(container.getNotes())
                                .usageContext(container.getUsageContext())
                                // Omits snippets for list request to reduce payload size.
                                // .headSnippet(snippetService.getHeadSnippet(container.getPublicId()))
                                // .bodySnippet(snippetService.getBodySnippet(container.getPublicId()))
                                .fingerprint(container.getFingerprint())
                                // Omits tags for list request to reduce payload size.
                                .tags(Collections.EMPTY_LIST)
                                .build();

                // Adds into the list;
                gtmContainers.add(gtmContainer);
            }

            log.info(
                    "listContainers for {} : {}",
                    accountResourceName,
                    JsonUtils.toJson(gtmContainers));
            return gtmContainers;
        } catch (IOException e) {
            log.error(e.getMessage());
            throw new DataException(e);
        }
    }

    private List<GTMTagDTO> listTags(String workspaceResourceName) throws DataException {
        try {
            List<GTMTagDTO> gtmTags = Lists.newArrayList();
            for (Tag tag :
                    Optional.ofNullable(
                                    tagManager
                                            .accounts()
                                            .containers()
                                            .workspaces()
                                            .tags()
                                            .list(workspaceResourceName)
                                            .execute()
                                            .getTag())
                            .orElse(List.of())) {
                GTMTagDTO gtmTag =
                        GTMTagDTO.builder()
                                .id(tag.getTagId())
                                .resourceName(tag.getPath())
                                .displayName(tag.getName())
                                .accountId(tag.getAccountId())
                                .containerId(tag.getContainerId())
                                .workspaceId(tag.getWorkspaceId())
                                .type(tag.getType())
                                .parameters(
                                        tag.getParameter().stream()
                                                .map(
                                                        parameter ->
                                                                GTMTagDTO.TagParameter.builder()
                                                                        .type(parameter.getType())
                                                                        .key(parameter.getKey())
                                                                        .value(parameter.getValue())
                                                                        .build())
                                                .toList())
                                .firingTriggerIds(tag.getFiringTriggerId())
                                .build();

                // Adds into the list;
                gtmTags.add(gtmTag);
            }

            log.info("listTags for {} : {}", workspaceResourceName, JsonUtils.toJson(gtmTags));
            return gtmTags;
        } catch (IOException e) {
            log.error(e.getMessage());
            throw new DataException(e);
        }
    }

    public GTMContainerDTO getContainer(String containerResourceName) throws DataException {
        try {
            Container container =
                    tagManager.accounts().containers().get(containerResourceName).execute();

            // Converts and returns.
            GTMContainerDTO gtmContainer =
                    GTMContainerDTO.builder()
                            .id(container.getContainerId())
                            .resourceName(container.getPath())
                            .displayName(container.getName())
                            .accountId(container.getAccountId())
                            .domainNames(container.getDomainName())
                            .publicId(container.getPublicId())
                            .tagIds(container.getTagIds())
                            .notes(container.getNotes())
                            .usageContext(container.getUsageContext())
                            .headSnippet(snippetService.getHeadSnippet(container.getPublicId()))
                            .bodySnippet(snippetService.getBodySnippet(container.getPublicId()))
                            .fingerprint(container.getFingerprint())
                            // Sets the tags associated with this container.
                            .tags(
                                    this.listTags(
                                            this.getDefaultWorkspace(container.getPath())
                                                    .getResourceName()))
                            .build();

            log.info("getContainer : {}", JsonUtils.toJson(gtmContainer));
            return gtmContainer;
        } catch (IOException e) {
            log.error(e.getMessage());
            throw new DataException(e);
        }
    }

    public GTMContainerDTO createContainer(String accountResourceName, String displayName)
            throws DataException {
        try {
            Container container = new Container();
            container.setName(displayName);
            container.setUsageContext(Collections.singletonList("web"));

            Container createdContainer =
                    tagManager
                            .accounts()
                            .containers()
                            .create(GTMAccountDTO.toResourceName(accountResourceName), container)
                            .execute();

            // Converts.
            GTMContainerDTO gtmContainer =
                    GTMContainerDTO.builder()
                            .id(createdContainer.getContainerId())
                            .resourceName(createdContainer.getPath())
                            .displayName(createdContainer.getName())
                            .accountId(parseId(accountResourceName))
                            .domainNames(createdContainer.getDomainName())
                            .publicId(createdContainer.getPublicId())
                            .tagIds(createdContainer.getTagIds())
                            .notes(createdContainer.getNotes())
                            .usageContext(createdContainer.getUsageContext())
                            .headSnippet(
                                    snippetService.getHeadSnippet(createdContainer.getPublicId()))
                            .bodySnippet(
                                    snippetService.getBodySnippet(createdContainer.getPublicId()))
                            .fingerprint(container.getFingerprint())
                            .build();

            log.info(
                    "createContainer for {} : {}",
                    accountResourceName,
                    JsonUtils.toJson(gtmContainer));
            return gtmContainer;
        } catch (IOException e) {
            log.error(e.getMessage());
            throw new DataException(e);
        }
    }

    /** Executes in async mode after the container is created to improve API efficiency. */
    @Async
    public void initContainer(
            String accountResourceName,
            String containerResourceName,
            String gaMeasurementId,
            String matomoSiteId)
            throws DataException {
        // Runs sequentially to prevent problems with API quota limitations.
        // See https://developers.google.com/tag-platform/tag-manager/api/v2/limits-quotas
        this.initContainerAccess(accountResourceName, containerResourceName);
        this.initContainerSetups(containerResourceName, gaMeasurementId, matomoSiteId);
    }

    private void initContainerAccess(String accountResourceName, String containerResourceName)
            throws DataException {
        try {
            // 1. Container access.
            // See
            // https://developers.google.com/tag-platform/tag-manager/api/v2/reference/accounts/user_permissions#containerAccess
            ContainerAccess containerAccess = new ContainerAccess();
            containerAccess.setContainerId(parseId(containerResourceName));
            // publish > approve > edit > read
            containerAccess.setPermission("publish");

            // 2. Account access - required in user permission request.
            // See
            // https://developers.google.com/tag-platform/tag-manager/api/v2/reference/accounts/user_permissions#accountAccess
            AccountAccess accountAccess = new AccountAccess();
            accountAccess.setPermission("admin");

            // 3. Combines as user permission.
            UserPermission userPermission = new UserPermission();
            userPermission.setContainerAccess(Collections.singletonList(containerAccess));
            userPermission.setAccountAccess(accountAccess);

            for (String containerAdmin : containerAdmins) {
                // One email address per request.
                userPermission.setEmailAddress(containerAdmin);
                UserPermission createdUserPermission =
                        tagManager
                                .accounts()
                                .userPermissions()
                                .create(accountResourceName, userPermission)
                                .execute();
                log.info("Created user permission {}", JsonUtils.toJson(createdUserPermission));
            }
        } catch (IOException e) {
            log.error(e.getMessage());
            throw new DataException(e);
        }
    }

    private void initContainerSetups(
            String containerResourceName, String gaMeasurementId, String matomoSiteId)
            throws DataException {
        GTMWorkspaceDTO gtmWorkspace = this.getDefaultWorkspace(containerResourceName);
        GTMTriggerDTO gtmTrigger = this.createTrigger(gtmWorkspace.getResourceName());
        this.createGATag(gtmWorkspace.getResourceName(), gtmTrigger.getId(), gaMeasurementId);
        if (StringUtils.isNotEmpty(matomoSiteId)) {
            this.createMatomoTag(
                    gtmWorkspace.getResourceName(),
                    gtmTrigger.getId(),
                    matomoTagService.getTrackingCode(matomoSiteId));
        }
        GTMVersionDTO gtmVersion = this.createContainerVersion(gtmWorkspace.getResourceName());
        this.publishContainerVersion(gtmVersion.getResourceName());
    }

    private GTMWorkspaceDTO getDefaultWorkspace(String containerResourceName) throws DataException {
        try {
            Workspace workspace =
                    tagManager
                            .accounts()
                            .containers()
                            .workspaces()
                            .list(containerResourceName)
                            .execute()
                            .getWorkspace()
                            .stream()
                            .filter(w -> w.getName().equalsIgnoreCase("Default Workspace"))
                            .findFirst()
                            .get();

            // Converts and returns.
            GTMWorkspaceDTO gtmWorkspace =
                    GTMWorkspaceDTO.builder()
                            .id(workspace.getWorkspaceId())
                            .resourceName(workspace.getPath())
                            .displayName(workspace.getName())
                            .accountId(workspace.getAccountId())
                            .containerId(workspace.getContainerId())
                            .description(workspace.getDescription())
                            .fingerprint(workspace.getFingerprint())
                            .build();

            log.info(
                    "getDefaultWorkspace for {} : {}",
                    containerResourceName,
                    JsonUtils.toJson(gtmWorkspace));
            return gtmWorkspace;
        } catch (IOException e) {
            log.error(e.getMessage());
            throw new DataException(e);
        }
    }

    private GTMTriggerDTO createTrigger(String workspaceResourceName) throws DataException {
        try {
            Trigger trigger = new Trigger();
            trigger.setName("Page View Trigger");
            // See
            // https://developers.google.com/tag-platform/tag-manager/api/v2/reference/accounts/containers/workspaces/triggers#type
            trigger.setType("pageview");

            Trigger createdTrigger =
                    tagManager
                            .accounts()
                            .containers()
                            .workspaces()
                            .triggers()
                            .create(workspaceResourceName, trigger)
                            .execute();

            // Converts and returns.
            GTMTriggerDTO gtmTrigger =
                    GTMTriggerDTO.builder()
                            .id(createdTrigger.getTriggerId())
                            .resourceName(createdTrigger.getPath())
                            .displayName(createdTrigger.getName())
                            .accountId(createdTrigger.getAccountId())
                            .containerId(createdTrigger.getContainerId())
                            .workspaceId(createdTrigger.getWorkspaceId())
                            .type(createdTrigger.getType())
                            .notes(createdTrigger.getNotes())
                            .fingerprint(createdTrigger.getFingerprint())
                            .build();

            log.info(
                    "createTrigger for {} : {}",
                    workspaceResourceName,
                    JsonUtils.toJson(gtmTrigger));
            return gtmTrigger;
        } catch (IOException e) {
            log.error(e.getMessage());
            throw new DataException(e);
        }
    }

    private void createGATag(String workspaceResourceName, String triggerId, String gaMeasurementId)
            throws DataException {
        try {
            // See
            // https://developers.google.com/tag-platform/tag-manager/api/v2/tag-dictionary-reference#ga-tag
            Parameter propertyParam = new Parameter();
            propertyParam.setType("template");
            propertyParam.setKey("tagId");
            propertyParam.setValue(gaMeasurementId);

            Parameter trackTypeParam = new Parameter();
            trackTypeParam.setType("template");
            trackTypeParam.setKey("trackType");
            trackTypeParam.setValue("TRACK_PAGEVIEW");

            Tag tag = new Tag();
            tag.setName("GA Tag - PV");
            tag.setType("googtag");
            tag.setLiveOnly(false);
            tag.setParameter(List.of(propertyParam, trackTypeParam));
            tag.setFiringTriggerId(Collections.singletonList(triggerId));

            Tag createdTag =
                    tagManager
                            .accounts()
                            .containers()
                            .workspaces()
                            .tags()
                            .create(workspaceResourceName, tag)
                            .execute();
            log.info("createGATag {}", JsonUtils.toJson(createdTag));
        } catch (IOException e) {
            log.error(e.getMessage());
            throw new DataException(e);
        }
    }

    private void createMatomoTag(
            String workspaceResourceName, String triggerId, String matomoTrackingCode)
            throws DataException {
        try {
            // See
            // https://developers.google.com/tag-platform/tag-manager/api/v2/tag-dictionary-reference#html
            Parameter htmlParam = new Parameter();
            htmlParam.setType("template");
            htmlParam.setKey("html");
            htmlParam.setValue(matomoTrackingCode);

            Parameter trackTypeParam = new Parameter();
            trackTypeParam.setType("template");
            trackTypeParam.setKey("trackType");
            trackTypeParam.setValue("TRACK_PAGEVIEW");

            Tag tag = new Tag();
            tag.setName("Matomo Tag - PV");
            tag.setType("html");
            tag.setLiveOnly(false);
            tag.setParameter(List.of(htmlParam, trackTypeParam));
            tag.setFiringTriggerId(Collections.singletonList(triggerId));

            Tag createdTag =
                    tagManager
                            .accounts()
                            .containers()
                            .workspaces()
                            .tags()
                            .create(workspaceResourceName, tag)
                            .execute();
            log.info("createMatomoTag {}", JsonUtils.toJson(createdTag));
        } catch (IOException e) {
            log.error(e.getMessage());
            throw new DataException(e);
        }
    }

    private GTMVersionDTO createContainerVersion(String workspaceResourceName)
            throws DataException {
        try {
            CreateContainerVersionRequestVersionOptions versionOptions =
                    new CreateContainerVersionRequestVersionOptions();
            versionOptions.setName(
                    "Latest version #"
                            + LocalDateTime.now()
                                    .format(DateTimeFormatter.ofPattern("yyyyMMddHHmmss")));

            ContainerVersion createdContainerVersion =
                    tagManager
                            .accounts()
                            .containers()
                            .workspaces()
                            .createVersion(workspaceResourceName, versionOptions)
                            .execute()
                            .getContainerVersion();

            // Converts and returns.
            GTMVersionDTO gtmVersion =
                    GTMVersionDTO.builder()
                            .id(createdContainerVersion.getContainerVersionId())
                            .resourceName(createdContainerVersion.getPath())
                            .displayName(createdContainerVersion.getName())
                            .accountId(createdContainerVersion.getAccountId())
                            .containerId(createdContainerVersion.getContainerId())
                            .deleted(BooleanUtils.isTrue(createdContainerVersion.getDeleted()))
                            .description(createdContainerVersion.getDescription())
                            .fingerprint(createdContainerVersion.getFingerprint())
                            .build();

            log.info(
                    "createContainerVersion for {} : {}",
                    workspaceResourceName,
                    JsonUtils.toJson(gtmVersion));
            return gtmVersion;
        } catch (IOException e) {
            log.error(e.getMessage());
            throw new DataException(e);
        }
    }

    private void publishContainerVersion(String versionResourceName) throws DataException {
        try {
            tagManager.accounts().containers().versions().publish(versionResourceName).execute();
            log.info("publishContainerVersion {}", versionResourceName);
        } catch (IOException e) {
            log.error(e.getMessage());
            throw new DataException(e);
        }
    }

    public void deleteContainer(String containerResourceName) throws DataException {
        try {
            tagManager.accounts().containers().delete(containerResourceName).execute();
            log.info("deleteContainer {}", containerResourceName);
        } catch (IOException e) {
            log.error(e.getMessage());
            throw new DataException(e);
        }
    }
}
