package com.wechi.adweb.bridge.matomo.controller;

import com.wechi.adweb.bridge.common.APIStatus;
import com.wechi.adweb.bridge.common.BaseController;
import com.wechi.adweb.bridge.common.OpenAPIRequest;
import com.wechi.adweb.bridge.common.OpenAPIResponse;
import com.wechi.adweb.bridge.exception.BadRequestException;
import com.wechi.adweb.bridge.matomo.dto.CreateSiteRequestDTO;
import com.wechi.adweb.bridge.matomo.dto.MatomoSiteDTO;
import com.wechi.adweb.bridge.matomo.service.MatomoService;
import com.wechi.adweb.bridge.util.JsonUtils;

import lombok.extern.slf4j.Slf4j;

import org.apache.commons.lang3.StringUtils;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.web.bind.annotation.*;

import java.util.List;

/**
 * @author wfansh
 */
@Slf4j
@RestController
@RequestMapping("/api/matomo")
public class MatomoController extends BaseController {
    @Autowired private MatomoService matomoService;

    @PostMapping("/sites/list")
    @ResponseBody
    public OpenAPIResponse<List<MatomoSiteDTO>> listSites(
            @RequestBody OpenAPIRequest<String> apiRequest) {
        long start = System.currentTimeMillis();
        log.info("****** listSites() ****** apiRequest = {}", JsonUtils.toJson(apiRequest));

        List<MatomoSiteDTO> matomoSites = matomoService.listSites();
        log.info("****** listSites() ****** duration = {} seconds", getElapsedSeconds(start));
        return OpenAPIResponse.<List<MatomoSiteDTO>>builder()
                .status(APIStatus.SUCCESS)
                .data(matomoSites)
                .build();
    }

    @PostMapping("/sites/get")
    @ResponseBody
    public OpenAPIResponse<MatomoSiteDTO> getSite(@RequestBody OpenAPIRequest<String> apiRequest)
            throws BadRequestException {
        long start = System.currentTimeMillis();
        log.info("****** getSite() ****** apiRequest = {}", JsonUtils.toJson(apiRequest));
        String siteId = apiRequest.getData();

        // 1. Validates the request parameters.
        if (StringUtils.isEmpty(siteId)) {
            throw new BadRequestException(apiRequest);
        }

        // 2. Executes the API request.
        MatomoSiteDTO matomoSite = matomoService.getSite(siteId);
        log.info("****** getSite() ****** duration = {} seconds", getElapsedSeconds(start));
        return OpenAPIResponse.<MatomoSiteDTO>builder()
                .status(APIStatus.SUCCESS)
                .data(matomoSite)
                .build();
    }

    @PostMapping("/sites/create")
    @ResponseBody
    public OpenAPIResponse<MatomoSiteDTO> createSite(
            @RequestBody OpenAPIRequest<CreateSiteRequestDTO> apiRequest)
            throws BadRequestException {
        long start = System.currentTimeMillis();
        log.info("****** createSite() ****** apiRequest = {}", JsonUtils.toJson(apiRequest));
        CreateSiteRequestDTO createSiteRequest = apiRequest.getData();

        // 1. Validates the request parameters.
        if (StringUtils.isEmpty(createSiteRequest.getSiteName())
                || StringUtils.isEmpty(createSiteRequest.getUrl())) {
            throw new BadRequestException(apiRequest);
        }

        // 2. Executes the API request.
        MatomoSiteDTO matomoSite =
                matomoService.createSite(
                        createSiteRequest.getSiteName(), createSiteRequest.getUrl());
        log.info("****** createSite() ****** duration = {} seconds", getElapsedSeconds(start));
        return OpenAPIResponse.<MatomoSiteDTO>builder()
                .status(APIStatus.SUCCESS)
                .data(matomoSite)
                .build();
    }

    @PostMapping("/sites/delete")
    @ResponseBody
    public OpenAPIResponse<String> deleteSite(@RequestBody OpenAPIRequest<String> apiRequest)
            throws BadRequestException {
        long start = System.currentTimeMillis();
        log.info("****** deleteSite() ****** apiRequest = {}", JsonUtils.toJson(apiRequest));
        String siteId = apiRequest.getData();

        // 1. Validates the request parameters.
        if (StringUtils.isEmpty(siteId)) {
            throw new BadRequestException(apiRequest);
        }

        // 2. Executes the API request.
        matomoService.deleteSite(siteId);
        log.info("****** deleteSite() ****** duration = {} seconds", getElapsedSeconds(start));
        return OpenAPIResponse.<String>builder().status(APIStatus.SUCCESS).data(siteId).build();
    }
}
