package com.wechi.adweb.bridge.matomo.service;

import com.google.common.collect.ImmutableMap;
import com.wechi.adweb.bridge.matomo.dto.MatomoSiteDTO;
import com.wechi.adweb.bridge.util.JsonUtils;

import lombok.extern.slf4j.Slf4j;

import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.core.ParameterizedTypeReference;
import org.springframework.http.RequestEntity;
import org.springframework.stereotype.Service;
import org.springframework.web.client.RestTemplate;
import org.springframework.web.util.UriComponentsBuilder;

import java.util.List;
import java.util.Map;

/**
 * @author wfansh
 *     <p>See https://matomous.adwebcloud.com/index.php?module=API&action=listAllAPI&idSite=12
 */
@Slf4j
@Service
public class MatomoService {

    private static final String QUERY_PARAM_MODULE = "module";
    private static final String QUERY_PARAM_FORMAT = "format";
    private static final String QUERY_PARAM_TOKEN_AUTH = "token_auth";
    private static final String QUERY_PARAM_METHOD = "method";
    private static final String QUERY_PARAM_SITE_ID = "idSite";
    private static final String QUERY_PARAM_SITE_NAME = "siteName";
    private static final String QUERY_PARAM_URLS = "urls";

    @Value("${matomo.url}")
    private String matomoUrl;

    @Value("${matomo.authToken}")
    private String matomoAuthToken;

    @Autowired private MatomoTagService matomoTagService;

    private final RestTemplate restTemplate = new RestTemplate();

    public List<MatomoSiteDTO> listSites() {
        List<MatomoSiteDTO> matomoSites =
                this.getForObject(
                        "SitesManager.getAllSites",
                        ImmutableMap.of(),
                        new ParameterizedTypeReference<>() {});

        // Omits the 'trackingCode' field for list request to reduce payload size.

        log.info("listSites : {}", JsonUtils.toJson(matomoSites));
        return matomoSites;
    }

    public MatomoSiteDTO getSite(String siteId) {
        MatomoSiteDTO matomoSite =
                this.getForObject(
                        "SitesManager.getSiteFromId",
                        ImmutableMap.of(QUERY_PARAM_SITE_ID, siteId),
                        ParameterizedTypeReference.forType(MatomoSiteDTO.class));

        // Adds additional fields.
        matomoSite.setTrackingCode(matomoTagService.getTrackingCode(matomoSite.getSiteId()));

        log.info("getSite : {}", JsonUtils.toJson(matomoSite));
        return matomoSite;
    }

    public MatomoSiteDTO createSite(String siteName, String url) {
        MatomoSiteDTO matomoSite =
                this.getForObject(
                        "SitesManager.addSite",
                        ImmutableMap.of(QUERY_PARAM_SITE_NAME, siteName, QUERY_PARAM_URLS, url),
                        ParameterizedTypeReference.forType(MatomoSiteDTO.class));

        // 'SitesManager.addSite' API method returns the 'value' field only, adds additional fields.
        matomoSite.setSiteName(siteName);
        matomoSite.setUrl(url);
        matomoSite.setTrackingCode(matomoTagService.getTrackingCode(matomoSite.getSiteId()));

        log.info("createSite : {}", JsonUtils.toJson(matomoSite));
        return matomoSite;
    }

    public void deleteSite(String siteId) {
        String result =
                this.getForObject(
                        "SitesManager.deleteSite",
                        ImmutableMap.of(QUERY_PARAM_SITE_ID, siteId),
                        ParameterizedTypeReference.forType(String.class));
        log.info("deleteSite {}, result = {}", siteId, result);
    }

    /** Sends a GET request to the Matomo API with the specified query parameters. */
    private <T> T getForObject(
            String apiMethod,
            Map<String, String> queryParams,
            ParameterizedTypeReference<T> responseType) {
        // 1. Builds the base URI with common API parameters.
        UriComponentsBuilder uriBuilder =
                UriComponentsBuilder.fromHttpUrl(matomoUrl)
                        .queryParam(QUERY_PARAM_METHOD, apiMethod)
                        .queryParam(QUERY_PARAM_MODULE, "API")
                        .queryParam(QUERY_PARAM_FORMAT, "json")
                        .queryParam(QUERY_PARAM_TOKEN_AUTH, matomoAuthToken);

        // 2. Adds specific API query parameters.
        queryParams.forEach(uriBuilder::queryParam);

        // 3. Executes the API request and returns the deserialized response.
        RequestEntity<?> requestEntity = RequestEntity.get(uriBuilder.build().toUri()).build();
        return restTemplate.exchange(requestEntity, responseType).getBody();
    }
}
